/***************************************************************************************************
*                            C# sample for the usage of class TopoART_AM                           *
****************************************************************************************************
*                            Created by Marko Tscherepanow, 9 March 2019                           *
***************************************************************************************************/

// Compile and run from the console: dotnet run -p TopoART-AM_sample1.csproj

using System;
using System.IO;
using System.Reflection;
using System.Text.RegularExpressions;
using LibTopoART;
using System.Globalization;

namespace LibTopoART_samples
{
	/// <summary>
	/// Sample using TopoART-AM with synthetic two-dimensional data. [C#]
	/// <para>
	/// A TopoART-AM network is trained with the well-known Two Spirals dataset augmented with additional information.
	/// The resulting network maps two-dimensional points lying on each spiral (key_1) to their Euclidean distance from 
	/// the origin and the corresponding spiral ID (key_2). Therefore, it can recall spiral points if a distance and a
	/// spiral ID are given, and vice versa.
	/// </para>
	/// <para>The resulting network can be visualised using the R script <c>ShowTopoARTAMResults.R</c> 
	/// provided in the subfolder <c>R</c>.
	/// </para>
	/// </summary>
	class TopoART_AM_sample1
	{
		static void Main() {

			// Dataset
			const string Dataset				=	"../../../../../data/TwoSpirals_dataset.txt";	
			// Destination directory for trained networks
			const string NetworkPath			=	"../../../../../results/networks/";

			const long SampleNumber				=	194;
			const long InterpolatedSampleNumber	=	20000;

			// Dimensions of the keys
			const long key_1_dimension = 2;
			const long key_2_dimension = 2;

			decimal[][] key_1_array = new decimal[SampleNumber + InterpolatedSampleNumber][];
			decimal[][] key_2_array = new decimal[SampleNumber + InterpolatedSampleNumber][];
			StreamReader DatasetFile;
			long i;
			string line;
			TopoART_AM tam;

			// Set working directory to assembly directory
            Directory.SetCurrentDirectory(Path.GetDirectoryName(new Uri(Assembly.GetEntryAssembly().CodeBase).LocalPath));

			// Load dataset and create keys (key_1: spiral points as two-dimensional coordinates; 
			// key_2: Euclidean distance from the origin and a numerical value reflecting the spiral ID)
			using(DatasetFile = new StreamReader(File.Open(Dataset, FileMode.Open))) {
				Random rnd = new Random();

				// Use original Two Spirals dataset
				for(i = 0; i < SampleNumber; ++i) {
					line = DatasetFile.ReadLine();
					string[] numbers = Regex.Split(line, @"\s+");

					key_1_array[i] = new decimal[key_1_dimension];
					// key_1_array[i] consists of the original coordinates of the two spirals dataset normalised to the interval [0, 1.0]
					key_1_array[i][0] = Decimal.Parse(numbers[1], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
					key_1_array[i][1] = Decimal.Parse(numbers[2], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
					key_1_array[i][0] = (key_1_array[i][0] + 7.0m) / 14m;
					key_1_array[i][1] = (key_1_array[i][1] + 7.0m) / 14m;

					key_2_array[i] = new decimal[key_2_dimension];
					// key_2_array[i][0] is computed as the Euclidean distance of key_1_array[i] from the centre of its input space (i.e., from [0.5, 0.5])
					key_2_array[i][0] = Math.Min((decimal)Math.Sqrt((double)((key_1_array[i][0] - 0.5m) * (key_1_array[i][0] - 0.5m) + (key_1_array[i][1] - 0.5m) * (key_1_array[i][1] - 0.5m))) * 2, 1.0m);
					// key_2_array[i][1] is set to a random value in [0.0, 0.1] for the first spiral and set to a random value in [0.9, 1.0] for the second spiral
					decimal ID_offset = (decimal)(rnd.NextDouble() * 0.1);
					key_2_array[i][1] = numbers[3].Equals("SPIRAL1") ? 0.0m + ID_offset : 1.0m - ID_offset;
				}

				// Add additional interpolated data, as otherwise TopoART-AM (like TopoART) cannot link the categories to clusters 
				// (Remark: The modified learning mechanism of Episodic TopoART does not require these interpolated data.)
				for(i = SampleNumber; i < SampleNumber + InterpolatedSampleNumber;) {
					decimal sample_offset = (decimal)(rnd.NextDouble() * (SampleNumber - 1));
					long start_index = (long)sample_offset;

					// Check maximum index and spiral ID
					if((start_index < SampleNumber - 1) && (Math.Round(key_2_array[start_index][1]) == Math.Round(key_2_array[start_index + 1][1]))) {
						decimal fraction = sample_offset - start_index;

						key_1_array[i] = new decimal[key_1_dimension];
						// Fill key_1_array[i] with additional interpolated values
						key_1_array[i][0] = fraction * key_1_array[start_index][0] + (1 - fraction) * key_1_array[start_index + 1][0];
						key_1_array[i][1] = fraction * key_1_array[start_index][1] + (1 - fraction) * key_1_array[start_index + 1][1];

						key_2_array[i] = new decimal[key_2_dimension];
						// key_2_array[i][0] is computed as the Euclidean distance of key_1_array[i] from the centre of its input space (i.e. from [0.5, 0.5])
						key_2_array[i][0] = Math.Min((decimal)Math.Sqrt((double)((key_1_array[i][0] - 0.5m) * (key_1_array[i][0] - 0.5m) + (key_1_array[i][1] - 0.5m) * (key_1_array[i][1] - 0.5m))) * 2, 1.0m);
						// key_2_array[i][1] is set to a random value in [0.0, 0.1] for the first spiral and set to a random value in [0.9, 1.0] for the second spiral
						decimal ID_offset = (decimal)(rnd.NextDouble() * 0.1);
						key_2_array[i][1] = (key_2_array[start_index][1] < 0.5m) ? 0.0m + ID_offset : 1.0m - ID_offset;

						++i;
					}
				}
			}

			// Initialise a TopoART-AM network with appropriate parameter values
			tam = new TopoART_AM(key_1_dimension, key_2_dimension, 2, 0.85m);
			tam.Beta_sbm	= 0.1m;
			tam.Phi			= 1;
			tam.Tau			= 200;

			// Train network
			for(i = 0; i < SampleNumber + InterpolatedSampleNumber; ++i)
				tam.Learn(key_1_array[i], key_2_array[i]);

			// Determine clusters
			tam.ComputeClusterIDs();

			// Save network in human readable form
			tam.SaveText(NetworkPath + "TopoART-AM_TwoSpirals_dataset.txt");

			// Save network in binary form
			tam.Save(NetworkPath + "TopoART-AM_TwoSpirals_dataset.tam");

			Console.WriteLine("");

			// Recall key_1 for a given key_2
			decimal[] recall_result;
			decimal F3_activation;
			decimal[] recall_key_2_vec = new decimal[key_2_dimension];
			recall_key_2_vec[0] = 0.25m;	// distance from the origin
			recall_key_2_vec[1] = 0.0m;		// spiral (0 to 0.1 = SPIRAL1, 0.9 to 1 = SPIRAL2)

			// Start recall process
			long recall_cluster_num = tam.BeginRecallKey1(recall_key_2_vec);
			Console.WriteLine("Consider " + recall_cluster_num + " clusters during recall");
			if(recall_cluster_num != 2) {
				// Such outcomes are possible due to the usage of random numbers. They do not constitute errors but
				// complicate the interpretation of the results shown below.
				Console.WriteLine("Warning: The expected cluster number differs from the actual one. Please repeat the training in case of any uncertainties.");
			}

			// The first recall step recalls the best result for the distance and spiral specified by recall_key_2_vec.
			if(tam.RecallStep(out recall_result, out F3_activation)) {
				Console.WriteLine("1st recall step: Approximated key_1 = (" + recall_result[0] + ", " + recall_result[1] + ")^T, " + "F3_Activation = " + F3_activation);
			}

			// The second recall step recalls the best result for the second cluster (i.e., the best fitting distance for the remaining spiral).
			// The reduced value of F3_activation reflects the dissimilarity caused by the association with the other spiral.
			if(tam.RecallStep(out recall_result, out F3_activation)) {
				Console.WriteLine("2nd recall step: Approximated key_1 = (" + recall_result[0] + ", " + recall_result[1] + ")^T, " + "F3_Activation = " + F3_activation);
			}

			// Finish recall process
			tam.EndRecall();	
		}
	}
}